#include <stdlib.h>
#include <cmath>
#include <iostream>
using std::cout;
using std::cerr;
using std::endl;
#include <fstream>
using std::ifstream;
using std::ofstream;

static double GRAD2DEG = 180.0 / M_PI;

inline double  rad(double grad) { return grad / GRAD2DEG; }
inline double grad(double rad)  { return  rad * GRAD2DEG; }
inline double  sqr(double x) { return x * x; }

// Klasse zur Darstellung einer 3x3-Matrix
class matrix3x3{  
public:
  double element[3][3];
  matrix3x3() { init(); }
  matrix3x3(const matrix3x3& b) {
    cout << "Copy Constructor" << endl;
    for (int i=0; i < 3; i++)
	 for (int j=0; j < 3; j++)
	   element[i][j] = b.element[i][j];    
  }  
  inline void init() {
    for (int i=0; i < 3; i++)
	 for (int j=0; j < 3; j++)
	   element[i][j] = 0.0;
  }
};

void print_matrix(const matrix3x3* M) {
  for (int i=0; i < 3; i++) {
    printf("   ( %4.2f  %4.2f  %10.8f )\n",
		 M->element[i][0], M->element[i][1],
		 M->element[i][2]);
  }
  printf("\n");
}

// Klasse zur Darstellung der Roboterpose (Zustand)
class pose {  
public:
  double x;
  double z;
  double theta;

  pose() { init(); }
  inline void init() { x = z = theta = 0.0; }
};


///////////////////////////////////////////////
//
// Implementationen einfacher Matrixoperationen
//
///////////////////////////////////////////////

void add_matrix(const matrix3x3* M1,
			 const matrix3x3* M2,
			 matrix3x3* M3)         // Addition der 3x3 Matrizen M1 und M2 zu M3
{
  for (int i=0; i < 3; i++)
    for (int j=0; j < 3; j++)
	 M3->element[i][j] = M1->element[i][j] + M2->element[i][j];
}

void sub_matrix(const matrix3x3* M1,
			 const matrix3x3* M2,
			 matrix3x3* M3)         // Subtraktion der 3x3 Matrizen M1 und M2 zu M3
{
  for (int i=0; i < 3; i++)
    for (int j=0; j < 3; j++)
	 M3->element[i][j] = M1->element[i][j] - M2->element[i][j];
}

void mult_matrix(const matrix3x3* M1,
			  const matrix3x3* M2,
			  matrix3x3* M3)        // Multiplikation der 3x3 Matrizen M1 und M2 zu M3
{
  int i,j;
  for(i=0;i<3;i++)
    {
	 for(j=0;j<3;j++)
	   {
		M3->element[i][j] =     M1->element[i][0] * M2->element[0][j]
		  + M1->element[i][1] * M2->element[1][j]
		  + M1->element[i][2] * M2->element[2][j];
	   }
    }
}

void mult_vector(const matrix3x3* M1,
			  const pose* P1,
			  pose* P2)
{
  P2->x     = P1->x * M1->element[0][0] + P1->z * M1->element[0][1] + P1->theta * M1->element[0][2];
  P2->z     = P1->x * M1->element[1][0] + P1->z * M1->element[1][1] + P1->theta * M1->element[1][2];
  P2->theta = P1->x * M1->element[2][0] + P1->z * M1->element[2][1] + P1->theta * M1->element[2][2];
}

void mult_scalar(const matrix3x3* M1, const double s, matrix3x3* M2)
{
  for (int i=0; i < 3; i++) {
    for (int j=0; j < 3; j++) {
      M2->element[i][j] = M1->element[i][j] * s;
    }
  }
}

void transp_matrix(matrix3x3* M1)      // Transposition einer 3x3 Matrix M1
{
  int i,j;
  matrix3x3 transp;
  for(i=0;i<3;i++)
    {
	 for(j=0;j<3;j++)
	   {
		transp.element[i][j] = M1->element[j][i];
	   }
    }
  for(i=0;i<3;i++)
    {
	 for(j=0;j<3;j++)
	   {
		M1->element[i][j] = transp.element[i][j];
	   }
    }
}


bool inverse_matrix(matrix3x3* M)      // Invertieren einer 3x3 Matrix M
{
  double det = 0.0;
  double a, b, c, d, e, f, g, h, i;
  a = M->element[0][0]; b = M->element[0][1]; c = M->element[0][2];
  d = M->element[1][0]; e = M->element[1][1]; f = M->element[1][2];
  g = M->element[2][0]; h = M->element[2][1]; i = M->element[2][2];
  det = g*(b*f-c*e) - h*(a*f-c*d) + i*(a*e-b*d);
  if (det == 0) { cerr << "DET = 0\n"; return false; }

  M->element[0][0] =  (e*i-f*h)/det;
  M->element[0][1] = -(b*i-c*h)/det;
  M->element[0][2] =  (b*f-c*e)/det;
  M->element[1][0] = -(d*i-f*g)/det;
  M->element[1][1] =  (a*i-c*g)/det;
  M->element[1][2] = -(a*f-c*d)/det;
  M->element[2][0] =  (d*h-e*g)/det;
  M->element[2][1] = -(a*h-b*g)/det;
  M->element[2][2] =  (a*e-b*d)/det;
  return true;
}

// pos1delta, pos2delta: Aktualisierungen der Pose (nur deltas!!!) von Odometry, Laser, ...
// *_err: Kovarianzmatrizen
// Ausgabe: Zustand (fusionierte Pose-detas), Sigma (= Sigma_t)
void measurement_update(const pose* pos1delta, const matrix3x3* pos1delta_err,
				    const pose* pos2delta, const matrix3x3* pos2delta_err,
				    pose* state, matrix3x3* Sigma)
{
  static matrix3x3 Sigma_bar; // Berechnung der Unsicherheit der berechneten Pose
  static pose tmpS, tmpS2;                        
  static matrix3x3 K, sigmaT, tmp, tmp2, tmp3;
  static matrix3x3 eins; 
  eins.element[0][0] = eins.element[1][1] = eins.element[2][2] = 1.0;
  
  // All dies ist unnoetig - sieht aber besser aus
  static matrix3x3 A = eins;
  static matrix3x3 H = eins;
  static matrix3x3 AT = A;
  transp_matrix(&AT);
  static matrix3x3 HT = H;
  transp_matrix(&HT);

  double d_x, d_z, d_theta;
  
  // a-priori-Vorhersage des Mittelwerts/Zustandes (vgl. Seite 137ff)
  
  /* todo?: Anwendung der Formel (4.50)? Muss hier etwas getan werden? Falls ja, hier oder wo sonst? */
  /* Nachdenken ....                                                                                 */

  
  // a-priori Kovarianz Schaetzung
  /* todo: a-priori-Vorhersage der Zustands-Kovarianz vgl. Seite 138 */
  
  // Kalman-Gewinn:
  /* todo: Formel (4.49) */
  
  
  // a-posteriori Kovarianz Schaetzung
  /* todo: Formel (4.48) */
  

  // a-posteriori Schätzung des Zustandes:
  //
  // Hinweis:  [Bitte nicht beachten, falls es fuer Sie keinen Sinn ergibt...]
  // --------
  // (d_x, d_z, d_theta) = pos2delta - pos1delta;  (== z - h(x) in Formel (4.51) )
  // Nur lokale Aenderungen => es ist moeglich zu subtrahieren
  //                           kein Wechsel des Koordinatensystems notwendig
  // Anschliessend: Multiplikation von K mit (d_x, d_z, d_theta)
  // Addition pos1delta, Speichern des resultierenden Zustandes

  /* todo: Formel (4.47) */
  
  cout << "==========================================================================================\n";

  cout << "Kalman-Gewinn:\n";
  print_matrix(&K);

  cout << "Pose Schätzung Kovarianz Matrix:\n";
  print_matrix(Sigma);
  return;
}

// Berechnung der Odometrie
void calcOdometry(double& o_vL, double& o_vR, pose* odom_delta) {
  static const double TURNING_ADAPTATION = 0.785;
  static const double AXIS_LENGTH        = 39.0;
  
  double dist = 0.5 * (o_vL + o_vR);
  odom_delta->theta  = (o_vL - o_vR) / AXIS_LENGTH * TURNING_ADAPTATION; 
  odom_delta->x      = dist * sin(odom_delta->theta);
  odom_delta->z      = dist * cos(odom_delta->theta);
  o_vL = o_vR = 0.0;  // reset accumulated values
}

// Transformation der pos_delta in das pos's Koordinatensystem, anschliessend
// Anwenden der Transformation (Addition)
void transform_coordinates(const pose* pos_delta, pose* pos) {
  pos->x += ( pos_delta->x * cos(pos->theta) - pos_delta->z * sin(pos->theta));
  pos->z += ( pos_delta->x * sin(pos->theta) + pos_delta->z * cos(pos->theta));
  pos->theta += pos_delta->theta;

  // Winkel normalisieren
  if (pos->theta >  M_PI ) pos->theta -= 2.0*M_PI;
  if (pos->theta < -M_PI)  pos->theta += 2.0*M_PI;
}

int main() {
  // Verwenden einer statische/konstanten Kovarianz-Matrix fuer laserlokalisierung (wie Odometrie)
  // oder verwenden der dynamisch berechneten matrix?
  bool useDynamicCovariance = false;
  
  /////////////////////////////////////////////////////////////
  //// [Odometrie Rauschen:] //////////////////////////////////

  // Aendern dieser Werte zur Modifikation der Kovarianz-Matrix der Odometrie
  double q1 = 0.5; // 1.0
  double q2 = 0.5; // 0.4
  double q3 = 0.5; // 0.5

  // Kovarianz-Matrizen
  matrix3x3 Q, R, K;  // Q: for odometry; R: laser localization; K: kalman
  
  // Laser Ruaschen:
  // Hier: statische Werte. Normalerweise nicht benutzt, da sich diese Kovarianz
  // berechnen laesst (vgl. Aufgabe 5.3)
  double r1 = 0.2;
  double r2 = 0.2;
  double r3 = 0.2;  

  // fuellen der Matrizen
  q3 = rad(q3); r3 = rad(r3);

  // Berechnung der Diagonalen <=> Matrix kann immer invertiert werden!
  Q.element[0][0] = q1*q1; Q.element[1][1] = q2*q2; Q.element[2][2] = q3*q3;
  if (!useDynamicCovariance) {
    R.element[0][0] = r1*r1; R.element[1][1] = r2*r2; R.element[2][2] = r3*r3;
  }
  
  // Posen
  pose odom_delta, laser_delta, kalman_delta, kalman_between_delta;
  pose odom, laser, kalman, kalman_between;
  
  ofstream out1("kalman_result.dat");
  ofstream out2("pose@scan.dat");  // zum Generieren von karten
  out1 << "#Format: Odometrie (x,z,t), Scanner (x,z,t), Kalman (x,z,t)\n";

  ifstream in_enc("encoder@scan.dat");
  ifstream in_scan("laser@scan.dat");
  if (!in_enc ) { cerr << "Error: encoder@scan.dat\n"; exit(1); }
  if (!in_scan) { cerr << "Error: laser@scan.dat\n"; exit(1); }

  int anz, z=0;
  double o_vL = 0, o_vR = 0, o_vL_step, o_vR_step, tmp;
  
  while (!(in_enc.eof() || in_scan.eof())) {
    z++;
    // Einlesen der Encoder-Werte
    in_enc >> o_vL_step >> o_vR_step;
    o_vL_step /= 10; o_vR_step /= 10;

    // Einlesen der Laserlokalisations-Werte
    in_scan >> laser_delta.x >> laser_delta.z >> laser_delta.theta;
    laser_delta.theta = rad(laser_delta.theta);  // Winkel ist in Grad im File gespeichert
    
    for (int j=0; j < 3; j++)
	 for (int k=0; k < 3; k++) {
	   in_scan >> tmp;
	   if (useDynamicCovariance) {
		// Nur Eintraege auf der Diagonalen; vgl. Q, R
		R.element[j][k] = ( (j != k) ? 0.0 : tmp);
	   }
	 }
    in_scan >> anz;  // Anzahl der gematchten Punkte

    // Akkumulieren der Odometireschritte.
    // ----------
    // Achtung:
    // Wir haben mehr Odometrieschritte als Laserscanmatchings
    // Die Files enthalten dennoch die gleiche Anzahl von Zeilen. Zeilen, die mit "-1"
    // enden sollten igonriert werden. 
    // Als ein Ergebnis haben wir die akkumulierten Odometrieschritte bis wir einen
    // korrekten Scan haben
    o_vL += o_vL_step; o_vR += o_vR_step;

    // verwende Scan, genau dann wenn anz != -1
    if (anz > 0) {

	 calcOdometry(o_vL, o_vR, &odom_delta);
	 
	 // ------- wird nicht benoetigt hier: nur zu Verlgeichszwecken:
	 transform_coordinates(&odom_delta, &odom);
	 transform_coordinates(&laser_delta, &laser);

	 // Anwendung des Filters:
	 measurement_update(&odom_delta, &Q, &laser_delta, &R, &kalman_delta, &K);
	 transform_coordinates(&kalman_delta, &kalman);

	 cout << z << ") "
		 << odom.x << " " << odom.z << " " << grad(odom.theta) << " | "
		 << laser.x << " " << laser.z << " " << grad(laser.theta) << " || "
		 << kalman.x << " " << kalman.z << " " << grad(kalman.theta) << endl;
	 
	 out1 << odom.x << " " << odom.z << " " << grad(odom.theta) << " "
		 << laser.x << " " << laser.z << " " << grad(laser.theta) << " "
		 << kalman.x << " " << kalman.z << " " << grad(kalman.theta) << endl;

	 out2 << kalman.x << " " << kalman.z << " " << kalman.theta << endl;
	 
	 kalman_between = kalman;  // Kopeiren der Pose
    }
    else { // falls nur odometrie-Aktualisierung, kein neues Scanmatching:
	 calcOdometry(o_vL_step, o_vR_step, &kalman_between_delta);	 
	 transform_coordinates(&kalman_between_delta, &kalman_between);

	 // Konsistenz mit Scan Matchingdaten (zum Erzeugen von Karten)
	 out2 << kalman_between.x << " " << kalman_between.z << " " << kalman_between.theta << endl;
    }
  }
  cout << endl << endl;
  cout << "> Files \"kalman_result.dat\" & \"pose@scan.dat\" (for bin/scan2map [k]) generated.\n\n";
}
